import textwrap
from typing import List, TYPE_CHECKING, Sequence

import pytest

from debputy.lsp.languages.lsp_debian_control import _reformat_debian_control
from debputy.lsp.maint_prefs import MaintainerPreferenceTable
from debputy.lsp.text_edit import apply_text_edits
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from lint_tests.lint_tutil import (
    ReformatWrapper,
    apply_formatting_edits,
)

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


ALL_REQUIRED_FILES = [
    "debian/changelog",
    "debian/copyright",
]


@pytest.fixture
def reformater(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lint_dctrl_parser: DctrlParser,
    maintainer_preference_table: MaintainerPreferenceTable,
) -> ReformatWrapper:

    return ReformatWrapper(
        "/nowhere/debian/control",
        _reformat_debian_control,
        debputy_plugin_feature_set,
        lint_dctrl_parser,
        maintainer_preference_table,
    )


def test_dctrl_reformat(reformater: ReformatWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Source: foo
    Some-Other-Field: bar
    Build-Depends: debhelper-compat (= 13)
    Homepage: https://example.com

    Package: foo
    Architecture: all
    Section: devel
    """
    ).splitlines(keepends=True)

    edits = reformater.reformat(lines)
    # By default, we do nothing
    assert not edits

    edits_black = reformater.reformat_with_named_style("black", lines)
    assert edits_black

    actual_reformatted_black = apply_formatting_edits(lines, edits_black)

    expected_reformatted_black = textwrap.dedent(
        """\
    Source: foo
    Some-Other-Field: bar
    Build-Depends:
     debhelper-compat (= 13),
    Homepage: https://example.com

    Package: foo
    Architecture: all
    Section: devel
    """
    )

    assert actual_reformatted_black == expected_reformatted_black


def test_dctrl_reformat_canonicalize_field_names(reformater: ReformatWrapper) -> None:
    lines = textwrap.dedent(
        """\
    source: foo
    # This field will not change case, since it is unknown
    some-other-field: bar
    BUILD-DEPENDS:
     debhelper-compat (= 13),
    # In memory of Steve (#1068220)
    HoMEpaGe: https://example.com

    package: foo
    architecture: all
    Section: devel
    """
    ).splitlines(keepends=True)

    edits = reformater.reformat(lines)
    # By default, we do nothing
    assert not edits

    edits_black = reformater.reformat_with_named_style("black", lines)
    assert edits_black

    actual_reformatted_black = apply_formatting_edits(lines, edits_black)

    expected_reformatted_black = textwrap.dedent(
        """\
    Source: foo
    # This field will not change case, since it is unknown
    some-other-field: bar
    Build-Depends:
     debhelper-compat (= 13),
    # In memory of Steve (#1068220)
    Homepage: https://example.com

    Package: foo
    Architecture: all
    Section: devel
    """
    )

    assert actual_reformatted_black == expected_reformatted_black


def test_dctrl_reformat_preserve_comments(reformater: ReformatWrapper) -> None:
    # Based on an example observed in  sbuild`. Thanks to zeha fore reporting.
    lines = textwrap.dedent(
        """\
    Source: foo
    # Test
    Build-Depends:
     foo,
     bar,

    Package: foo
    # Test
    architecture: i386 amd64
    Section: devel
    # Test
    Depends: foo
    # Test
    recommends:
     bar,
    """
    ).splitlines(keepends=True)

    edits = reformater.reformat(lines)
    # By default, we do nothing
    assert not edits

    edits_black = reformater.reformat_with_named_style("black", lines)
    assert edits_black

    actual_reformatted_black = apply_formatting_edits(lines, edits_black)

    expected_reformatted_black = textwrap.dedent(
        """\
    Source: foo
    # Test
    Build-Depends:
     bar,
     foo,

    Package: foo
    # Test
    Architecture: amd64 i386
    Section: devel
    # Test
    Depends:
     foo,
    # Test
    Recommends:
     bar,
    """
    )

    assert actual_reformatted_black == expected_reformatted_black



import dataclasses
from typing import List, Optional, FrozenSet, TYPE_CHECKING

from debputy.lsp.lsp_features import SEMANTIC_TOKENS_LEGEND
from debputy.util import grouper

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


@dataclasses.dataclass(slots=True, frozen=True)
class ResolvedSemanticToken:
    range: types.Range
    value: str
    token_name: str
    modifiers: FrozenSet[str] = frozenset()


def resolved_semantic_token(
    line_no: int,
    col_start: int,
    value: str,
    token_type: str,
    *,
    token_modifiers: FrozenSet[str] = frozenset(),
) -> ResolvedSemanticToken:
    return ResolvedSemanticToken(
        types.Range(
            types.Position(
                line_no,
                col_start,
            ),
            types.Position(
                line_no,
                col_start + len(value),
            ),
        ),
        value,
        token_type,
        token_modifiers,
    )


def resolve_semantic_tokens(
    lines: List[str],
    token_result: Optional[types.SemanticTokens],
) -> Optional[List[ResolvedSemanticToken]]:
    if token_result is None:
        return None
    assert (len(token_result.data) % 5) == 0
    current_line = 0
    current_col = 0
    resolved_tokens = []
    token_types = SEMANTIC_TOKENS_LEGEND.token_types
    for token_data in grouper(token_result.data, 5, incomplete="strict"):
        line_delta, col_start_delta, token_len, token_code, modifier_codes = token_data
        if line_delta:
            current_col = 0
        current_line += line_delta
        current_col += col_start_delta
        assert (
            not modifier_codes
        ), "TODO: Modifiers not supported (no modifiers defined)"

        value = lines[current_line][current_col : current_col + token_len]

        resolved_tokens.append(
            resolved_semantic_token(
                current_line,
                current_col,
                value,
                token_types[token_code],
            ),
        )

    return resolved_tokens



import dataclasses
from typing import (
    TypeVar,
    Optional,
    Union,
    List,
    Callable,
    TYPE_CHECKING,
    Sequence,
    Mapping,
    FrozenSet,
)

from debputy.manifest_parser.exceptions import ManifestTypeException
from debputy.packages import BinaryPackage
from debputy.util import assume_not_none

if TYPE_CHECKING:
    from debputy.manifest_parser.util import AttributePath
    from debputy.manifest_parser.parser_data import ParserContextData

S = TypeVar("S")
T = TypeVar("T")


@dataclasses.dataclass(slots=True, frozen=True)
class PackageSelectorRule:
    valid_values: FrozenSet[str]
    impl: Callable[[BinaryPackage, str], bool]


def _pkg_selector_arch(package: BinaryPackage, selection_value: str) -> bool:
    if selection_value == "all":
        return package.is_arch_all
    return not package.is_arch_all


PACKAGE_SELECTORS: Mapping[str, PackageSelectorRule] = {
    "arch": PackageSelectorRule(
        frozenset(["all", "any"]),
        _pkg_selector_arch,
    ),
    "package-type": PackageSelectorRule(
        frozenset(["deb", "udeb"]),
        lambda p, v: p.package_type == v,
    ),
}


def type_mapper_str2package(
    raw_package_name: str,
    ap: "AttributePath",
    opc: Optional["ParserContextData"],
) -> BinaryPackage:
    pc = assume_not_none(opc)
    if "{{" in raw_package_name:
        resolved_package_name = pc.substitution.substitute(raw_package_name, ap.path)
    else:
        resolved_package_name = raw_package_name

    package_name_in_message = raw_package_name
    if resolved_package_name != raw_package_name:
        package_name_in_message = f'"{resolved_package_name}" ["{raw_package_name}"]'

    if not pc.is_known_package(resolved_package_name):
        package_names = ", ".join(pc.binary_packages)
        raise ManifestTypeException(
            f'The value {package_name_in_message} (from "{ap.path}") does not reference a package declared in'
            f" debian/control. Valid package names are: {package_names}"
        )
    package_data = pc.binary_package_data(resolved_package_name)
    if package_data.is_auto_generated_package:
        package_names = ", ".join(pc.binary_packages)
        raise ManifestTypeException(
            f'The package name {package_name_in_message} (from "{ap.path}") references an auto-generated package.'
            " However, auto-generated packages are now permitted here. Valid package names are:"
            f" {package_names}"
        )
    return package_data.binary_package


@dataclasses.dataclass(slots=True, frozen=True)
class PackageSelector:
    matched_binary_packages: Sequence[BinaryPackage]

    @classmethod
    def parse(
        cls,
        raw_value: str,
        attribute_path: "AttributePath",
        parser_context: "ParserContextData",
    ) -> "PackageSelector":
        pc = assume_not_none(parser_context)
        if ":" in raw_value:
            key, value = raw_value.split(":", maxsplit=1)
            selector = PACKAGE_SELECTORS.get(key)
            if selector is None:
                package_names = ", ".join(pc.binary_packages)
                valid_selectors = ", ".join(
                    f"{k}:<value>" for k in sorted(PACKAGE_SELECTORS)
                )
                raise ManifestTypeException(
                    f'Unknown package selector "{raw_value}" (from "{attribute_path.path}").'
                    f" Valid values here are: {package_names}, OR one of {valid_selectors}"
                )

            if value not in selector.valid_values:
                valid_values = ", ".join(sorted(selector.valid_values))
                raise ManifestTypeException(
                    f'The selection criteria "{key}" does not accept "{value}" (from "{attribute_path.path}").'
                    f" Valid values here are: {valid_values}"
                )

            matches = tuple(
                p
                for p in parser_context.binary_packages.values()
                if selector.impl(p, value)
            )
            return PackageSelector(matches)
        package = type_mapper_str2package(raw_value, attribute_path, parser_context)
        return PackageSelector((package,))


def wrap_into_list(
    x: T,
    _ap: "AttributePath",
    _pc: Optional["ParserContextData"],
) -> List[T]:
    return [x]


def normalize_into_list(
    x: Union[T, List[T]],
    _ap: "AttributePath",
    _pc: Optional["ParserContextData"],
) -> List[T]:
    return x if isinstance(x, list) else [x]


def map_each_element(
    mapper: Callable[[S, "AttributePath", Optional["ParserContextData"]], T],
) -> Callable[[List[S], "AttributePath", Optional["ParserContextData"]], List[T]]:
    def _generated_mapper(
        xs: List[S],
        ap: "AttributePath",
        pc: Optional["ParserContextData"],
    ) -> List[T]:
        return [mapper(s, ap[i], pc) for i, s in enumerate(xs)]

    return _generated_mapper



import re
from email.utils import parsedate_to_datetime
from typing import Optional, Tuple

from debputy.linting.lint_util import LintState
from debputy.lsp.lsp_features import (
    lsp_standard_handler,
    SecondaryLanguage,
    LanguageDispatchRule,
    lint_diagnostics,
)
from debputy.lsp.quickfixes import (
    propose_correct_text_quick_fix,
)
from debputy.lsp.spellchecking import spellcheck_line
from debputy.lsprotocol.types import (
    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
    TEXT_DOCUMENT_CODE_ACTION,
)
from debputy.util import PKGVERSION_REGEX

try:
    from debputy.lsp.vendoring._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )

    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass


# Same as Lintian
_MAXIMUM_WIDTH: int = 82
_HEADER_LINE = re.compile(r"^(\S+)\s*[(]([^)]+)[)]")  # TODO: Add rest
_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/changelog",
    None,
    ("debian/changelog", "debian/changelog.dch"),
    [
        # emacs's name
        SecondaryLanguage("debian-changelog"),
        # vim's name
        SecondaryLanguage("debchangelog"),
        SecondaryLanguage("dch"),
    ],
)


_WEEKDAYS_BY_IDX = [
    "Mon",
    "Tue",
    "Wed",
    "Thu",
    "Fri",
    "Sat",
    "Sun",
]
_KNOWN_WEEK_DAYS = frozenset(_WEEKDAYS_BY_IDX)


lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)

DPM_DCH_SECTION = "Policy 4.4"


def _check_footer_date(
    lint_state: LintState,
    line: str,
    line_no: int,
    line_len: int,
    start_date_idx: int,
) -> None:
    # 3 characters for the day name (Mon), then a comma plus a space followed by the
    # actual date. The 6 characters limit is a gross under estimation of the real
    # size.
    if line_len < start_date_idx + 6:
        text_range = _single_line_subrange(line_no, start_date_idx, line_len)
        lint_state.emit_diagnostic(
            text_range,
            "Expected a date in RFC822 format (Tue, 12 Mar 2024 12:34:56 +0000)",
            "error",
            DPM_DCH_SECTION,
        )
        return
    day_name_range = _single_line_subrange(line_no, start_date_idx, start_date_idx + 3)
    day_name = line[start_date_idx : start_date_idx + 3]
    if day_name not in _KNOWN_WEEK_DAYS:
        lint_state.emit_diagnostic(
            day_name_range,
            "Expected a three letter date here using US English format (Mon, Tue, ..., Sun)",
            "error",
            DPM_DCH_SECTION,
        )
        return

    date_str = line[start_date_idx + 5 :]

    if line[start_date_idx + 3 : start_date_idx + 5] != ", ":
        sep = line[start_date_idx + 3 : start_date_idx + 5]
        text_range = _single_line_subrange(
            line_no,
            start_date_idx + 3,
            start_date_idx + 5,
        )
        lint_state.emit_diagnostic(
            text_range,
            f'Improper formatting of date. Expected ", " here, not "{sep}"',
            "error",
            DPM_DCH_SECTION,
        )
        return

    try:
        # FIXME: this parser is too forgiving (it ignores trailing garbage)
        date = parsedate_to_datetime(date_str)
    except ValueError as e:
        error_range = _single_line_subrange(line_no, start_date_idx + 5, line_len)
        lint_state.emit_diagnostic(
            error_range,
            f"Unable to parse the date as a valid RFC822 date: {e.args[0]}",
            "error",
            "debputy",
        )
        return
    expected_week_day = _WEEKDAYS_BY_IDX[date.weekday()]
    if expected_week_day != day_name:
        lint_state.emit_diagnostic(
            day_name_range,
            f"The date was a {expected_week_day}day",
            "warning",
            "debputy",
            quickfixes=[propose_correct_text_quick_fix(expected_week_day)],
        )


def _check_email(
    lint_state: LintState,
    line: str,
    line_no: int,
    line_len: int,
) -> Tuple[int, int]:
    email_start_idx = _offset_of(line, "<", 4)
    email_end_idx = _offset_of(line, ">", max(email_start_idx, 4))

    if not (3 < email_start_idx < email_end_idx):
        # Email invalid
        if email_start_idx >= 3:
            msg = 'Missing closing ">" to finish email address before the sign off date'
            diag_start = email_start_idx
            diag_end = _offset_of(
                line,
                " ",
                email_start_idx,
                offset_if_missing=line_len,
            )
        else:
            if email_end_idx > -1:
                diag_start = 1 + _roffset_of(
                    line,
                    " ",
                    4,
                    email_end_idx - 1,
                    offset_if_missing=4,
                )
                if diag_start > 4:
                    email_start_idx = diag_start
                # For consistency, we always include the trailing `>`.
                diag_end = email_end_idx + 1
            else:
                diag_start = 4
                diag_end = _offset_of(
                    line,
                    "  ",
                    diag_start,
                    offset_if_missing=line_len,
                )
            msg = 'Missing opening "<" to start the email address after the name'
        lint_state.emit_diagnostic(
            _single_line_subrange(line_no, diag_start, diag_end),
            msg,
            "error",
            DPM_DCH_SECTION,
        )
        if email_end_idx < 0:
            email_end_idx = diag_end
    return email_start_idx, email_end_idx


def _check_footer_line(
    lint_state: LintState,
    line: str,
    line_no: int,
) -> None:
    if line.rstrip() == " --":
        lint_state.emit_diagnostic(
            _single_line_subrange(line_no, 0, 3),
            'Missing "Name <email@example.com>"',
            "error",
            DPM_DCH_SECTION,
            # TODO: Add quick fix to insert all the relevant data.
        )
        return
    line_len = len(line)
    if not line.startswith(" -- "):
        # Pre-condition for this function being called.
        assert line.startswith(" --") and line_len > 3
        lint_state.emit_diagnostic(
            _single_line_subrange(line_no, 0, line_len),
            'Start of sign-off line should be " -- ".',
            "error",
            DPM_DCH_SECTION,
            quickfixes=[propose_correct_text_quick_fix(" -- ")],
        )
        return

    email_start_idx, email_end_idx = _check_email(
        lint_state,
        line,
        line_no,
        line_len,
    )

    start_of_email_identified = email_start_idx > 3
    end_of_email_identified = 4 < email_end_idx < line_len
    if not start_of_email_identified:
        return

    # Email valid, name might be missing
    name_start = 4
    name_end = email_start_idx
    name = line[name_start:name_end]
    if not name or name.isspace():
        # The range must always be at least one character width. There is no good direction
        # to expand it in. Picked left because we know there is a space character,
        if name_end - name_start < 1:
            name_start -= 1
        lint_state.emit_diagnostic(
            _single_line_subrange(line_no, name_start, name_end),
            "Missing name before email",
            "error",
            DPM_DCH_SECTION,
        )
    elif not name.endswith(" ") or len(name.strip()) != len(name) - 1:
        lint_state.emit_diagnostic(
            _single_line_subrange(line_no, name_start, name_end),
            "Non-standard spacing around the name",
            "warning",
            DPM_DCH_SECTION,
            quickfixes=[
                propose_correct_text_quick_fix(
                    name.strip() + " ",
                    proposed_title="Fix spacing",
                )
            ],
        )
    if not end_of_email_identified:
        # If we are unsure where the email is, we currently do not have the wherewithal to identify
        # where the date is. Technically, there are cases where we could identify the date and work
        # back from there. Not written because I thought it is too much effort for the value at the
        # time I put in this comment.
        #
        # Note this will already have triggered a diagnostic.
        return
    post_email = line[email_end_idx + 1 :]
    if not post_email or post_email.isspace():
        lint_state.emit_diagnostic(
            _single_line_subrange(line_no, 0, line_len),
            "Missing sign off date",
            "error",
            DPM_DCH_SECTION,
        )
        return
    start_date_idx = email_end_idx + (len(post_email) - len(post_email.lstrip())) + 1
    space_len = start_date_idx - email_end_idx - 1
    if line[email_end_idx + 1 : start_date_idx] != "  ":
        correction = "  "
        diag_start = email_end_idx + 1
        diag_end = start_date_idx
        if not space_len:
            # If there is no spaces, then we mark the closing `>` and the following character instead if possible.
            #
            # Note the choice here of including both boundary characters is because some editors refuse to include
            # the lone `>` in a codeAction range with the user explicitly doing a selection range, so our range
            # detection will miss it. By including the following character, we ensure there is always a two
            # character range to place the cursor in the middle of and the editors tend to respect that as a valid
            # range (also, single character ranges are harder for the user to see).
            diag_start = email_end_idx
            diag_end = min(start_date_idx + 1, line_len)
            if start_date_idx < line_len:
                end_char = line[start_date_idx]
            else:
                end_char = ""
            correction = f">  {end_char}"

        lint_state.emit_diagnostic(
            _single_line_subrange(
                line_no,
                diag_start,
                diag_end,
            ),
            "Must be exactly two spaces between email and sign off date",
            "error",
            DPM_DCH_SECTION,
            quickfixes=[
                propose_correct_text_quick_fix(
                    correction,
                    proposed_title="Fix spacing between email and date",
                ),
            ],
        )
    _check_footer_date(lint_state, line, line_no, line_len, start_date_idx)


def _offset_of(
    text: str,
    ch: str,
    /,
    start: Optional[int] = None,
    end: Optional[int] = None,
    *,
    offset_if_missing: int = -1,
) -> int:
    try:
        return text.index(ch, start, end)
    except ValueError:
        return offset_if_missing


def _roffset_of(
    text: str,
    ch: str,
    /,
    start: Optional[int] = None,
    end: Optional[int] = None,
    *,
    offset_if_missing: int = -1,
) -> int:
    try:
        return text.rindex(ch, start, end)
    except ValueError:
        return offset_if_missing


def _single_line_subrange(
    line_no: int,
    character_start_pos: int,
    character_end_pos: int,
) -> "TERange":
    return TERange(
        TEPosition(
            line_no,
            character_start_pos,
        ),
        TEPosition(
            line_no,
            character_end_pos,
        ),
    )


def _check_header_line(
    lint_state: LintState,
    line: str,
    line_no: int,
    entry_no: int,
) -> None:
    m = _HEADER_LINE.search(line)
    if not m:
        # Syntax error: TODO flag later
        return
    source_name, source_version = m.groups()
    dctrl_source_pkg = lint_state.source_package
    if (
        entry_no == 1
        and dctrl_source_pkg is not None
        and dctrl_source_pkg.fields.get("Source") != source_name
    ):
        expected_name = dctrl_source_pkg.fields.get("Source")
        start_pos, end_pos = m.span(1)
        name_range = _single_line_subrange(line_no, start_pos, end_pos)
        if expected_name is None:
            msg = (
                "The first entry must use the same source name as debian/control."
                ' The d/control file is missing the "Source" field in its first stanza'
            )
        else:
            msg = (
                "The first entry must use the same source name as debian/control."
                f' Changelog uses: "{source_name}" while d/control uses: "{expected_name}"'
            )

        lint_state.emit_diagnostic(
            name_range,
            msg,
            "error",
            "dpkg",  # man:deb-src-control(5) / #1089794
        )
    if not PKGVERSION_REGEX.fullmatch(source_version):
        vm = PKGVERSION_REGEX.search(source_version)
        start_pos, end_pos = m.span(2)
        if vm:
            start_valid, end_valid = vm.span(0)
            invalid_ranges = []
            if start_valid > 0:
                name_range = _single_line_subrange(
                    line_no,
                    start_pos,
                    start_pos + start_valid,
                )
                invalid_ranges.append(name_range)

            if end_valid < len(source_version):
                name_range = _single_line_subrange(
                    line_no,
                    start_pos + end_valid,
                    end_pos,
                )
                invalid_ranges.append(name_range)

            for r in invalid_ranges:
                lint_state.emit_diagnostic(
                    r,
                    "This part cannot be parsed as a valid Debian version",
                    "error",
                    "Policy 5.6.12",
                )
        else:
            name_range = _single_line_subrange(line_no, start_pos, end_pos)
            lint_state.emit_diagnostic(
                name_range,
                f'Cannot parse "{source_version}" as a Debian version.',
                "error",
                "Policy 5.6.12",
            )
    elif "dsfg" in source_version:
        typo_index = source_version.index("dsfg")
        start_pos, end_pos = m.span(2)

        name_range = _single_line_subrange(
            line_no,
            start_pos + typo_index,
            start_pos + typo_index + 4,
        )
        lint_state.emit_diagnostic(
            name_range,
            'Typo of "dfsg" (Debian Free Software Guidelines)',
            "pedantic",
            "debputy",
            quickfixes=[propose_correct_text_quick_fix("dfsg")],
        )


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_changelog(lint_state: LintState) -> None:
    lines = lint_state.lines
    entry_no = 0
    entry_limit = 2
    max_words = 1000
    max_line_length = _MAXIMUM_WIDTH
    for line_no, line in enumerate(lines):
        orig_line = line
        line = line.rstrip()
        if not line:
            continue
        if line.startswith(" --"):
            _check_footer_line(lint_state, line, line_no)
            continue
        if not line.startswith("  "):
            if not line[0].isspace():
                entry_no += 1
                # Figure out the right cut which may not be as simple as just the
                # top two.
                if entry_no > entry_limit:
                    break
                _check_header_line(
                    lint_state,
                    line,
                    line_no,
                    entry_no,
                )
            continue
        # minus 1 for newline
        orig_line_len = len(orig_line) - 1
        if orig_line_len > max_line_length:
            exceeded_line_range = _single_line_subrange(
                line_no,
                max_line_length,
                orig_line_len,
            )
            lint_state.emit_diagnostic(
                exceeded_line_range,
                f"Line exceeds {max_line_length} characters",
                "pedantic",
                "debputy",
            )
        if len(line) > 3 and line[2] == "[" and line[-1] == "]":
            # Do not spell check [ X ] as X is usually a name
            continue
        if max_words > 0:
            new_diagnostics = spellcheck_line(lint_state, line_no, line)
            max_words -= new_diagnostics



import re
from typing import Union, Callable, Optional, Iterable, Iterator, Any

from debputy.exceptions import PluginInitializationError
from debputy.plugin.api import VirtualPath
from debputy.plugin.api.impl import DebputyPluginInitializerProvider
from debputy.plugin.api.spec import (
    DebputyPluginDefinition,
    PackageProcessor,
    PackageTypeSelector,
)


class ExperimentalDebputyPluginDefinition(DebputyPluginDefinition):
    """Plugin definition entity for plugins that use experimental API

    The plugin definition provides a way for the plugin code to register the features it provides via
    accessors. This particular instance also provides accessors for experimental or yet-to-be-final
    APIs. These APIs generally requires the plugin to be pre-registered with `debputy`, so `debputy`
    can add `Breaks` on the plugin provider if/when the API changes.
    """

    def package_processor(
        self,
        func: Optional[PackageProcessor] = None,
        *,
        processor_id: Optional[str] = None,
        depends_on_processor: Iterable[str] = tuple(),
        package_type: PackageTypeSelector = "deb",
    ) -> Union[
        Callable[[PackageProcessor], PackageProcessor],
        PackageProcessor,
    ]:
        """Provide a pre-assembly hook that can affect the contents of the binary ("deb") packages

        The provided hook will be run once per binary package to be assembled, and it can see all the content
        ("data.tar") planned to be included in the deb. It has read-write access to the data contents.

        The hook will be run unconditionally for all binary packages built. When the hook does not apply to all
        packages, it must provide its own (internal) logic for detecting whether it is relevant and reduce itself
        to a no-op if it should not apply to the current package.

        Hooks between plugins run in "some implementation defined order" and should not rely on being run before
        or after any other hook. However, the hooks can depend on other hooks from the same plugin by using
        the `depends_on_processor`.

        The hooks are only applied to packages defined in `debian/control`. Notably, the processor will
        not apply to auto-generated `-dbgsym` packages (as those are not listed explicitly in `debian/control`).

        >>> plugin_definition = define_debputy_plugin_experimental_api()
        >>> def _la_files(fs_root: VirtualPath) -> Iterator[VirtualPath]:
        ...     ...  # Implementation skipped here for brevity
        >>>
        >>> @plugin_definition.package_processor
        ... def clean_la_files(
        ...     fs_root: "VirtualPath",
        ...     _: Any,  # Not yet defined/provided
        ...     context: "PackageProcessingContext",
        ... ) -> None:
        ...     for path in _la_files(fs_root):
        ...         buffer = []
        ...         with path.open(byte_io=True) as fd:
        ...             replace_file = False
        ...             for line in fd:
        ...                 if line.startswith(b"dependency_libs"):
        ...                     replacement = re.sub(b"<real_pattern_here>", b"''", line)
        ...                     if replacement != line:
        ...                         replace_file = True
        ...                         line = replacement
        ...                 buffer.append(line)
        ...             if not replace_file:
        ...                 continue
        ...             with path.replace_fs_path_content() as fs_path, open(fs_path, "wb") as wfd:
        ...                 wfd.writelines(buffer)

        :param func: The function to be decorated/registered.
        :param processor_id: A plugin-wide unique ID for this detector. Packagers may use this ID for disabling
          the detector and accordingly the ID is part of the plugin's API toward the packager.
        :param depends_on_processor: Denote processors from the same plugin that must run before this processor.
          It is not possible to depend on processors from other plugins.
        :param package_type: Which kind of packages this metadata detector applies to.  The package type is generally
          defined by `Package-Type` field in the binary package. The default is to only run for regular `deb` packages
          and ignore `udeb` packages.
        """

        def _decorate(f: PackageProcessor) -> PackageProcessor:

            final_id = self._name2id(processor_id, f.__name__)

            def _init(api: "DebputyPluginInitializerProvider") -> None:
                api.package_processor(
                    final_id,
                    f,
                    depends_on_processor=depends_on_processor,
                    package_type=package_type,
                )

            self._generic_initializers.append(_init)

            return f

        if func:
            return _decorate(func)
        return _decorate

    def initialize(self, api: "DebputyPluginInitializerProvider") -> None:
        """Initialize the plugin from this definition

        Most plugins will not need this function as the plugin loading will call this function
        when relevant. However, a plugin can call this manually from a function-based initializer.
        This is mostly useful if the function-based initializer need to set up a few features
        that the `DebputyPluginDefinition` cannot do and the mix/match approach is not too
        distracting for plugin maintenance.

        :param api: The plugin initializer provided by the `debputy` plugin system
        """

        initializers = self._generic_initializers
        if not initializers:
            plugin_name = api.plugin_metadata.plugin_name
            raise PluginInitializationError(
                f"Initialization of {plugin_name}: The plugin definition was never used to register any features."
                " If you want to conditionally register features, please use an initializer functon instead."
            )

        for initializer in initializers:
            initializer(api)


def define_debputy_plugin_experimental_api() -> ExperimentalDebputyPluginDefinition:
    return ExperimentalDebputyPluginDefinition()
